@@ -23,21 +23,27 @@ L.Socket = L.Class.extend({
 
 	connect: function() {
 		var map = this._map;
+		if (map.options.permission) {
+			map.options.docParams['permission'] = map.options.permission;
+		}
+		var wopiSrc = '';
+		if (map.options.wopiSrc != '') {
+			wopiSrc = '?WOPISrc=' + map.options.wopiSrc + '&compat=/ws';
+		}
+
 		try {
-			if (map.options.permission) {
-				map.options.docParams['permission'] = map.options.permission;
-			}
-			this.socket = new WebSocket(map.options.server + '/lool/' + encodeURIComponent(map.options.doc + '?' + $.param(map.options.docParams)) + '/ws');
-			this.socket.onerror = L.bind(this._onSocketError, this);
-			this.socket.onclose = L.bind(this._onSocketClose, this);
-			this.socket.onopen = L.bind(this._onSocketOpen, this);
-			this.socket.onmessage = L.bind(this._onMessage, this);
-			this.socket.binaryType = 'arraybuffer';
+			this.socket = new WebSocket(map.options.server + '/lool/' + encodeURIComponent(map.options.doc + '?' + $.param(map.options.docParams)) + '/ws' + wopiSrc);
 		} catch (e) {
 			this._map.fire('error', {msg: _('Oops, there is a problem connecting to LibreOffice Online : ' + e), cmd: 'socket', kind: 'failed', id: 3});
-			return null;
+			return;
 		}
 
+		this.socket.onerror = L.bind(this._onSocketError, this);
+		this.socket.onclose = L.bind(this._onSocketClose, this);
+		this.socket.onopen = L.bind(this._onSocketOpen, this);
+		this.socket.onmessage = L.bind(this._onMessage, this);
+		this.socket.binaryType = 'arraybuffer';
+
 		if (map.options.docParams.access_token && parseInt(map.options.docParams.access_token_ttl)) {
 			var tokenExpiryWarning = 900 * 1000; // Warn when 15 minutes remain
 			clearTimeout(this._accessTokenExpireTimeout);
@@ -66,6 +72,9 @@ L.Socket = L.Class.extend({
 		this.socket.onmessage = function () {};
 		this.socket.close();
 
+		// Reset wopi's app loaded so that reconnecting again informs outerframe about initialization
+		this._map['wopi'].resetAppLoaded();
+		this._map.fire('docloaded', {status: false});
 		clearTimeout(this._accessTokenExpireTimeout);
 	},
 
@@ -82,7 +91,7 @@ L.Socket = L.Class.extend({
 
 		var socketState = this.socket.readyState;
 		if (socketState === 2 || socketState === 3) {
-			this.initialize(this._map);
+			this._map.loadDocument();
 		}
 
 		if (socketState === 1) {
@@ -113,9 +122,17 @@ L.Socket = L.Class.extend({
 		this.socket.send(msg);
 	},
 
+	_getParameterByName: function(url, name) {
+		name = name.replace(/[\[]/, '\\[').replace(/[\]]/, '\\]');
+		var regex = new RegExp('[\\?&]' + name + '=([^&#]*)'), results = regex.exec(url);
+		return results === null ? '' : results[1].replace(/\+/g, ' ');
+	},
+
 	_onSocketOpen: function () {
 		console.debug('_onSocketOpen:');
 		this._map._serverRecycling = false;
+		this._map._documentIdle = false;
+
 		// Always send the protocol version number.
 		// TODO: Move the version number somewhere sensible.
 		this._doSend('loolclient ' + this.ProtocolVersionNumber);
@@ -225,23 +242,44 @@ L.Socket = L.Class.extend({
 			this._map.fire('wopiprops', wopiInfo);
 			return;
 		}
+		else if (textMsg.startsWith('commandresult: ')) {
+			var commandresult = JSON.parse(textMsg.substring(textMsg.indexOf('{')));
+			if (commandresult['command'] === 'savetostorage' && commandresult['success']) {
+				// Close any open confirmation dialogs
+				if (vex.dialogID > 0) {
+					var id = vex.dialogID;
+					vex.dialogID = -1;
+					vex.close(id);
+				}
+			}
+			return;
+		}
 		else if (textMsg.startsWith('close: ')) {
 			textMsg = textMsg.substring('close: '.length);
 			msg = '';
-
+			var postMsgData = {};
 			// This is due to document owner terminating the session
 			if (textMsg === 'ownertermination') {
 				msg = _('Session terminated by document owner');
+				postMsgData['Reason'] = 'OwnerTermination';
 			}
-			else if (textMsg === 'idle') {
-				msg = _('Session terminated due to idleness');
+			else if (textMsg === 'idle' || textMsg === 'oom') {
+				msg = _('Idle document - please click to reload and resume editing');
+				this._map._documentIdle = true;
+				postMsgData['Reason'] = 'DocumentIdle';
+				if (textMsg === 'oom')
+					postMsgData['Reason'] = 'OOM';
 			}
 			else if (textMsg === 'shuttingdown') {
 				msg = _('Server is shutting down for maintenance (auto-saving)');
+				postMsgData['Reason'] = 'ShuttingDown';
+			}
+			else if (textMsg === 'docdisconnected') {
+				msg = _('Oops, there is a problem connecting the document');
+				postMsgData['Reason'] = 'DocumentDisconnected';
 			}
 			else if (textMsg === 'recycling') {
 				msg = _('Server is recycling and will be available shortly');
-
 				this._map._active = false;
 				this._map._serverRecycling = true;
 
@@ -265,11 +303,32 @@ L.Socket = L.Class.extend({
 					}
 
 					try {
-						socket.initialize(map);
+						map.loadDocument(map);
 					} catch (error) {
 					}
 				}, timeoutMs);
 			}
+			else if (textMsg.startsWith('documentconflict')) {
+				msg = _('Document has changed in storage. Loading the new document. Your version is available as revision.');
+
+				if (this._map._docLayer) {
+					this._map._docLayer.removeAllViews();
+				}
+				// Detach all the handlers from current socket, otherwise _onSocketClose tries to reconnect again
+				// However, we want to reconnect manually here.
+				this.close();
+
+				// Reload the document
+				this._map._active = false;
+				map = this._map;
+				vex.timer = setInterval(function() {
+					try {
+						// Activate and cancel timer and dialogs.
+						map._activate();
+					} catch (error) {
+					}
+				}, 3000);
+			}
 
 			// Close any open dialogs first.
 			if (vex.dialogID > 0) {
@@ -278,12 +337,17 @@ L.Socket = L.Class.extend({
 				vex.close(id);
 			}
 
+			var message = '';
+			if (!this._map['wopi'].DisableInactiveMessages) {
+				message = msg;
+			}
+
 			var options = $.extend({}, vex.defaultOptions, {
 				contentCSS: {'background':'rgba(0, 0, 0, 0)',
 				             'font-size': 'xx-large',
 				             'color': '#fff',
 				             'text-align': 'center'},
-				content: msg
+				content: message
 			});
 			options.id = vex.globalID;
 			vex.dialogID = options.id;
@@ -302,21 +366,31 @@ L.Socket = L.Class.extend({
 			});
 			options.$vex.append(options.$vexContent);
 
+			if (textMsg === 'idle' || textMsg === 'oom') {
+				var map = this._map;
+				options.$vex.bind('click.vex', function(e) {
+					console.debug('idleness: reactivating');
+					map._documentIdle = false;
+					return map._activate();
+				});
+			}
+
 			$(options.appendLocation).append(options.$vex);
 			vex.setupBodyClassName(options.$vex);
 
-			if (textMsg !== 'shuttingdown') {
+			if (postMsgData['Reason']) {
 				// Tell WOPI host about it which should handle this situation
-				this._map.fire('postMessage', {msgId: 'Session_Closed'});
+				this._map.fire('postMessage', {msgId: 'Session_Closed', args: postMsgData});
 			}
 
-			if (textMsg === 'idle' || textMsg === 'ownertermination') {
+			if (textMsg === 'ownertermination') {
 				this._map.remove();
 			}
 
 			return;
 		}
 		else if (textMsg.startsWith('error:') && command.errorCmd === 'storage') {
+			this._map.hideBusy();
 			var storageError;
 			if (command.errorKind === 'savediskfull') {
 				storageError = errorMessages.storage.savediskfull;
@@ -332,9 +406,40 @@ L.Socket = L.Class.extend({
 				// Since this is a document load failure, wsd will disconnect the socket anyway,
 				// better we do it first so that another error message doesn't override this one
 				// upon socket close.
-				this._map.hideBusy();
 				this.close();
 			}
+			else if (command.errorKind === 'documentconflict')
+			{
+				storageError = errorMessages.storage.documentconflict;
+
+				// TODO: We really really need to factor this out duplicate dialog code logic everywhere
+				// Close any open dialogs first.
+				if (vex.dialogID > 0) {
+					var id = vex.dialogID;
+					vex.dialogID = -1;
+					vex.close(id);
+				}
+
+				vex.dialog.confirm({
+					message: _('Document has been changed in storage. Do you want to discard your changes, and load the document from the storage?'),
+					buttons: [
+						$.extend({}, vex.dialog.buttons.YES, { text: _('Discard changes') }),
+						$.extend({}, vex.dialog.buttons.NO, { text: _('Overwrite document') })
+					],
+					callback: L.bind(function(value) {
+						if (value) {
+							// They want to refresh the page and load document again for all
+							this.sendMessage('closedocument');
+						} else {
+							// They want to overwrite
+							this.sendMessage('savetostorage force=1');
+						}
+					}, this)
+				});
+				vex.dialogID = vex.globalID - 1;
+
+				return;
+			}
 
 			// Parse the storage url as link
 			var tmpLink = document.createElement('a');
@@ -346,12 +451,12 @@ L.Socket = L.Class.extend({
 			return;
 		}
 		else if (textMsg.startsWith('error:') && command.errorCmd === 'internal') {
+			this._map.hideBusy();
 			this._map._fatal = true;
 			if (command.errorKind === 'diskfull') {
 				this._map.fire('error', {msg: errorMessages.diskfull});
 			}
 			else if (command.errorKind === 'unauthorized') {
-				this._map.hideBusy();
 				this._map.fire('error', {msg: errorMessages.unauthorized});
 			}
 
@@ -362,7 +467,11 @@ L.Socket = L.Class.extend({
 
 			return;
 		}
+		else if (textMsg.startsWith('error:') && command.errorCmd === 'saveas') {
+			this._map.hideBusy();
+		}
 		else if (textMsg.startsWith('error:') && command.errorCmd === 'load') {
+			this._map.hideBusy();
 			this.close();
 
 			var errorKind = command.errorKind;
@@ -412,10 +521,10 @@ L.Socket = L.Class.extend({
 					callback: L.bind(function(data) {
 						if (data) {
 							this._map._docPassword = data.password;
-							this.initialize(this._map);
+							this._map.loadDocument();
 						} else if (passwordType === 'to-modify') {
 							this._map._docPassword = '';
-							this.initialize(this._map);
+							this._map.loadDocument();
 						} else {
 							this._map.hideBusy();
 						}
@@ -459,6 +568,32 @@ L.Socket = L.Class.extend({
 					', last: ' + (command.rendercount - this._map._docLayer._debugRenderCount));
 			this._map._docLayer._debugRenderCount = command.rendercount;
 		}
+		else if (textMsg.startsWith('saveas:')) {
+			this._map.hideBusy();
+			if (command !== undefined && command.url !== undefined && command.url !== '') {
+				this.close();
+				var url = command.url;
+				var accessToken = this._getParameterByName(url, 'access_token');
+				var accessTokenTtl = this._getParameterByName(url, 'access_token_ttl');
+
+				if (accessToken !== undefined) {
+					if (accessTokenTtl === undefined) {
+						accessTokenTtl = 0;
+					}
+					this._map.options.docParams = { 'access_token': accessToken, 'access_token_ttl': accessTokenTtl };
+				}
+				else {
+					this._map.options.docParams = {};
+				}
+
+				// setup for loading the new document, and trigger the load
+				var docUrl = url.split('?')[0];
+				this._map.options.doc = docUrl;
+				this._map.options.wopiSrc = encodeURIComponent(docUrl);
+				this._map.loadDocument();
+			}
+			// var name = command.name; - ignored, we get the new name via the wopi's BaseFileName
+		}
 		else if (textMsg.startsWith('statusindicator:')) {
 			//FIXME: We should get statusindicator when saving too, no?
 			this._map.showBusy(_('Connecting...'), false);
@@ -501,59 +636,64 @@ L.Socket = L.Class.extend({
 			var img = 'data:image/png;base64,' + window.btoa(strBytes);
 		}
 
-		if (textMsg.startsWith('status:') && !this._map._docLayer) {
-			// first status message, we need to create the document layer
-			var tileWidthTwips = this._map.options.tileWidthTwips;
-			var tileHeightTwips = this._map.options.tileHeightTwips;
-			if (this._map.options.zoom !== this._map.options.defaultZoom) {
-				var scale = this._map.options.crs.scale(this._map.options.defaultZoom - this._map.options.zoom);
-				tileWidthTwips = Math.round(tileWidthTwips * scale);
-				tileHeightTwips = Math.round(tileHeightTwips * scale);
-			}
-
-			var docLayer = null;
-			if (command.type === 'text') {
-				docLayer = new L.WriterTileLayer('', {
-					permission: this._map.options.permission,
-					tileWidthTwips: tileWidthTwips,
-					tileHeightTwips: tileHeightTwips,
-					docType: command.type
-				});
-			}
-			else if (command.type === 'spreadsheet') {
-				docLayer = new L.CalcTileLayer('', {
-					permission: this._map.options.permission,
-					tileWidthTwips: tileWidthTwips,
-					tileHeightTwips: tileHeightTwips,
-					docType: command.type
-				});
-			}
-			else {
-				if (command.type === 'presentation' &&
-						this._map.options.defaultZoom === this._map.options.zoom) {
-					// If we have a presentation document and the zoom level has not been set
-					// in the options, resize the document so that it fits the viewing area
-					var verticalTiles = this._map.getSize().y / 256;
-					tileWidthTwips = Math.round(command.height / verticalTiles);
-					tileHeightTwips = Math.round(command.height / verticalTiles);
+		if (textMsg.startsWith('status:')) {
+			if (!this._map._docLayer) {
+				// first status message, we need to create the document layer
+				var tileWidthTwips = this._map.options.tileWidthTwips;
+				var tileHeightTwips = this._map.options.tileHeightTwips;
+				if (this._map.options.zoom !== this._map.options.defaultZoom) {
+					var scale = this._map.options.crs.scale(this._map.options.defaultZoom - this._map.options.zoom);
+					tileWidthTwips = Math.round(tileWidthTwips * scale);
+					tileHeightTwips = Math.round(tileHeightTwips * scale);
 				}
-				docLayer = new L.ImpressTileLayer('', {
-					permission: this._map.options.permission,
-					tileWidthTwips: tileWidthTwips,
-					tileHeightTwips: tileHeightTwips,
-					docType: command.type
-				});
+
+				var docLayer = null;
+				if (command.type === 'text') {
+					docLayer = new L.WriterTileLayer('', {
+						permission: this._map.options.permission,
+						tileWidthTwips: tileWidthTwips,
+						tileHeightTwips: tileHeightTwips,
+						docType: command.type
+					});
+				}
+				else if (command.type === 'spreadsheet') {
+					docLayer = new L.CalcTileLayer('', {
+						permission: this._map.options.permission,
+						tileWidthTwips: tileWidthTwips,
+						tileHeightTwips: tileHeightTwips,
+						docType: command.type
+					});
+				}
+				else {
+					if (command.type === 'presentation' &&
+					    this._map.options.defaultZoom === this._map.options.zoom) {
+						// If we have a presentation document and the zoom level has not been set
+						// in the options, resize the document so that it fits the viewing area
+						var verticalTiles = this._map.getSize().y / 256;
+						tileWidthTwips = Math.round(command.height / verticalTiles);
+						tileHeightTwips = Math.round(command.height / verticalTiles);
+					}
+					docLayer = new L.ImpressTileLayer('', {
+						permission: this._map.options.permission,
+						tileWidthTwips: tileWidthTwips,
+						tileHeightTwips: tileHeightTwips,
+						docType: command.type
+					});
+				}
+
+				this._map._docLayer = docLayer;
+				this._map.addLayer(docLayer);
+				this._map.fire('doclayerinit');
+			}
+			else if (this._reconnecting) {
+				// we are reconnecting ...
+				this._reconnecting = false;
+				this._map._docLayer._onMessage('invalidatetiles: EMPTY', null);
+				this._map.fire('statusindicator', {statusType: 'reconnected'});
+				this._map.setPermission(this._map.options.permission);
 			}
 
-			this._map._docLayer = docLayer;
-			this._map.addLayer(docLayer);
-			this._map.fire('doclayerinit');
-		} else if (textMsg.startsWith('status:') && this._reconnecting) {
-			// we are reconnecting ...
-			this._reconnecting = false;
-			this._map._docLayer._onMessage('invalidatetiles: EMPTY', null);
-			this._map.fire('statusindicator', {statusType: 'reconnected'});
-			this._map.setPermission(this._map.options.permission);
+			this._map.fire('docloaded', {status: true});
 		}
 
 		// these can arrive very early during the startup
@@ -598,8 +738,9 @@ L.Socket = L.Class.extend({
 			this._map.fire('error', {msg: _('Well, this is embarrassing, we cannot connect to your document. Please try again.'), cmd: 'socket', kind: 'closed', id: 4});
 		}
 
-		// Reset wopi's app loaded so that reconnecting again informs outerframe about initialization again
+		// Reset wopi's app loaded so that reconnecting again informs outerframe about initialization
 		this._map['wopi'].resetAppLoaded();
+		this._map.fire('docloaded', {status: false});
 
 		if (!this._reconnecting) {
 			this._reconnecting = true;
@@ -679,6 +820,9 @@ L.Socket = L.Class.extend({
 			else if (tokens[i].substring(0, 5) === 'char=') {
 				command.char = tokens[i].substring(5);
 			}
+			else if (tokens[i].substring(0, 4) === 'url=') {
+				command.url = tokens[i].substring(4);
+			}
 			else if (tokens[i].substring(0, 7) === 'viewid=') {
 				command.viewid = tokens[i].substring(7);
 			}
@@ -691,8 +835,8 @@ L.Socket = L.Class.extend({
 			else if (tokens[i].substring(0, 12) === 'rendercount=') {
 				command.rendercount = parseInt(tokens[i].substring(12));
 			}
-			else if (tokens[i].startsWith('hash=')) {
-				command.hash = this.getParameterValue(tokens[i]);
+			else if (tokens[i].startsWith('wid=')) {
+				command.wireId = this.getParameterValue(tokens[i]);
 			}
 		}
 		if (command.tileWidth && command.tileHeight && this._map._docLayer) {
