@@ -339,7 +339,7 @@ L.GridLayer = L.Layer.extend({
 			this._tileZoom = tileZoom;
 			if (tileZoomChanged) {
 				this._updateTileTwips();
-				this._updateMaxBounds();
+				this._updateMaxBounds(null, null, zoom);
 			}
 			this._updateLevels();
 			this._resetGrid();
@@ -363,16 +363,19 @@ L.GridLayer = L.Layer.extend({
 		this._tileHeightTwips = Math.round(this.options.tileHeightTwips * factor);
 	},
 
-	_updateMaxBounds: function (sizeChanged, extraSize, options) {
+	_updateMaxBounds: function (sizeChanged, extraSize, options, zoom) {
 		if (this._docWidthTwips === undefined || this._docHeightTwips === undefined) {
 			return;
 		}
+		if (!zoom) {
+			var zoom = this._map.getZoom();
+		}
 		var docPixelLimits = new L.Point(this._docWidthTwips / this.options.tileWidthTwips,
 			this._docHeightTwips / this.options.tileHeightTwips);
 		docPixelLimits = extraSize ? docPixelLimits.multiplyBy(this._tileSize).add(extraSize) :
 			docPixelLimits.multiplyBy(this._tileSize);
 
-		var scale = this._map.getZoomScale(this._map.getZoom(), 10);
+		var scale = this._map.getZoomScale(zoom, 10);
 		var topLeft = new L.Point(0, 0);
 		topLeft = this._map.unproject(topLeft.multiplyBy(scale));
 		var bottomRight = new L.Point(docPixelLimits.x, docPixelLimits.y);
@@ -538,13 +541,7 @@ L.GridLayer = L.Layer.extend({
 				// so we're able to cancel the previous requests that are being processed
 				this._map._socket.sendMessage('canceltiles');
 				for (key in this._tiles) {
-					// When _invalidCount > 0 the tile has been invalidated, however the new tile content
-					// has not yet been fetched and because of `canceltiles` message it will never be
-					// so we need to remove the tile, or when the tile is back inside the visible area
-					// its content would be the old invalidated one;
-					// example: a tile is invalidated but a sudden scroll to the cell cursor position causes
-					// to move the tile out of the visible area before the new content is fetched
-					if (!this._tiles[key].loaded || this._tiles[key]._invalidCount > 0) {
+					if (!this._tiles[key].loaded) {
 						L.DomUtil.remove(this._tiles[key].el);
 						delete this._tiles[key];
 						if (this._debug) {
@@ -617,13 +614,7 @@ L.GridLayer = L.Layer.extend({
 				this._map._socket.sendMessage('canceltiles');
 				for (key in this._tiles) {
 					tile = this._tiles[key];
-					// When _invalidCount > 0 the tile has been invalidated, however the new tile content
-					// has not yet been fetched and because of `canceltiles` message it will never be
-					// so we need to remove the tile, or when the tile is back inside the visible area
-					// its content would be the old invalidated one;
-					// example: a tile is invalidated but a sudden scroll to the cell cursor position causes
-					// to move the tile out of the visible area before the new content is fetched
-					if (!tile.loaded || tile._invalidCount > 0) {
+					if (!tile.loaded) {
 						L.DomUtil.remove(tile.el);
 						delete this._tiles[key];
 						if (this._debug && this._debugDataCancelledTiles) {
