@@ -14,16 +14,17 @@
 #include <fstream>
 
 #include <Poco/Net/HTTPResponse.h>
+#include <Poco/StringTokenizer.h>
 #include <Poco/URI.h>
 
-#include "Common.hpp"
 #include "DocumentBroker.hpp"
 #include "LOOLWSD.hpp"
-#include "Log.hpp"
-#include "Protocol.hpp"
-#include "Session.hpp"
-#include "Util.hpp"
-#include "Unit.hpp"
+#include "common/Common.hpp"
+#include "common/Log.hpp"
+#include "common/Protocol.hpp"
+#include "common/Session.hpp"
+#include "common/Unit.hpp"
+#include "common/Util.hpp"
 
 using namespace LOOLProtocol;
 
@@ -39,7 +40,8 @@ ClientSession::ClientSession(const std::string& id,
     _uriPublic(uriPublic),
     _isDocumentOwner(false),
     _isAttached(false),
-    _isViewLoaded(false)
+    _isViewLoaded(false),
+    _keyEvents(1)
 {
     const size_t curConnections = ++LOOLWSD::NumConnections;
     LOG_INF("ClientSession ctor [" << getName() << "], current number of connections: " << curConnections);
@@ -64,7 +66,7 @@ bool ClientSession::_handleInput(const char *buffer, int length)
 {
     LOG_TRC(getName() << ": handling incoming [" << getAbbreviatedMessage(buffer, length) << "].");
     const std::string firstLine = getFirstLine(buffer, length);
-    const auto tokens = LOOLProtocol::tokenize(firstLine.data(), firstLine.size());
+    const std::vector<std::string> tokens = LOOLProtocol::tokenize(firstLine.data(), firstLine.size());
 
     auto docBroker = getDocumentBroker();
     if (!docBroker)
@@ -81,9 +83,14 @@ bool ClientSession::_handleInput(const char *buffer, int length)
         updateLastActivityTime();
         docBroker->updateLastActivityTime();
     }
-
     if (tokens[0] == "loolclient")
     {
+        if (tokens.size() < 1)
+        {
+            sendTextFrame("error: cmd=loolclient kind=badprotocolversion");
+            return false;
+        }
+
         const auto versionTuple = ParseVersion(tokens[1]);
         if (std::get<0>(versionTuple) != ProtocolMajorVersionNumber ||
             std::get<1>(versionTuple) != ProtocolMinorVersionNumber)
@@ -135,6 +142,7 @@ bool ClientSession::_handleInput(const char *buffer, int length)
              tokens[0] != "resetselection" &&
              tokens[0] != "save" &&
              tokens[0] != "saveas" &&
+             tokens[0] != "savetostorage" &&
              tokens[0] != "selectgraphic" &&
              tokens[0] != "selecttext" &&
              tokens[0] != "setclientpart" &&
@@ -172,6 +180,13 @@ bool ClientSession::_handleInput(const char *buffer, int length)
             LOG_DBG("Session [" << getId() << "] requested owner termination");
             docBroker->closeDocument("ownertermination");
         }
+        else if (docBroker->isDocumentChangedInStorage())
+        {
+            LOG_DBG("Document marked as changed in storage and user ["
+                    << getUserId() << ", " << getUserName()
+                    << "] wants to refresh the document for all.");
+            docBroker->closeDocument("documentconflict " + getUserName());
+        }
 
         return true;
     }
@@ -209,12 +224,30 @@ bool ClientSession::_handleInput(const char *buffer, int length)
     {
         int dontTerminateEdit = 1;
         int dontSaveIfUnmodified = 1;
-        getTokenInteger(tokens[1], "dontTerminateEdit", dontTerminateEdit);
-        getTokenInteger(tokens[2], "dontSaveIfUnmodified", dontSaveIfUnmodified);
+        if (tokens.size() > 1)
+            getTokenInteger(tokens[1], "dontTerminateEdit", dontTerminateEdit);
+
+        if (tokens.size() > 2)
+            getTokenInteger(tokens[2], "dontSaveIfUnmodified", dontSaveIfUnmodified);
+
         docBroker->sendUnoSave(getId(), dontTerminateEdit != 0, dontSaveIfUnmodified != 0);
     }
+    else if (tokens[0] == "savetostorage")
+    {
+        int force = 0;
+        if (tokens.size() > 1)
+            getTokenInteger(tokens[1], "force", force);
+
+        if (docBroker->saveToStorage(getId(), true, "" /* This is irrelevant when success is true*/, true))
+        {
+            docBroker->broadcastMessage("commandresult: { \"command\": \"savetostorage\", \"success\": true }");
+        }
+    }
     else
     {
+        if (tokens[0] == "key")
+            _keyEvents++;
+
         if (!filterMessage(firstLine))
         {
             const std::string dummyFrame = "dummymsg";
@@ -260,11 +293,18 @@ bool ClientSession::loadDocument(const char* /*buffer*/, int /*length*/,
         {
             std::string encodedUserId;
             Poco::URI::encode(_userId, "", encodedUserId);
-            oss << " authorid=" + encodedUserId;
+            oss << " authorid=" << encodedUserId;
 
             std::string encodedUserName;
             Poco::URI::encode(_userName, "", encodedUserName);
-            oss << " author=" + encodedUserName;
+            oss << " author=" << encodedUserName;
+        }
+
+        if (!_userExtraInfo.empty())
+        {
+            std::string encodedUserExtraInfo;
+            Poco::URI::encode(_userExtraInfo, "", encodedUserExtraInfo);
+            oss << " authorextrainfo=" << encodedUserExtraInfo;
         }
 
         oss << " readonly=" << isReadOnly();
@@ -284,6 +324,13 @@ bool ClientSession::loadDocument(const char* /*buffer*/, int /*length*/,
             oss << " lang=" << _lang;
         }
 
+        if (!_watermarkText.empty())
+        {
+            std::string encodedWatermarkText;
+            Poco::URI::encode(_watermarkText, "", encodedWatermarkText);
+            oss << " watermarkText=" << encodedWatermarkText;
+        }
+
         if (!_docOptions.empty())
         {
             oss << " options=" << _docOptions;
@@ -497,7 +544,7 @@ bool ClientSession::handleKitToClientMessage(const char* buffer, const int lengt
 {
     const auto payload = std::make_shared<Message>(buffer, length, Message::Dir::Out);
 
-    LOG_TRC(getName() + ": handling kit-to-client [" << payload->abbr() << "].");
+    LOG_TRC(getName() << ": handling kit-to-client [" << payload->abbr() << "].");
     const std::string& firstLine = payload->firstLine();
 
     const auto docBroker = _docBroker.lock();
@@ -574,38 +621,74 @@ bool ClientSession::handleKitToClientMessage(const char* buffer, const int lengt
         int curPart;
         return getTokenInteger(tokens[1], "part", curPart);
     }
-    else if (tokens.size() == 2 && tokens[0] == "saveas:")
+    else if (tokens.size() == 3 && tokens[0] == "saveas:")
     {
-        std::string url;
-        if (!getTokenString(tokens[1], "url", url))
+        bool isConvertTo = static_cast<bool>(_saveAsSocket);
+
+        std::string encodedURL;
+        if (!getTokenString(tokens[1], "url", encodedURL))
         {
             LOG_ERR("Bad syntax for: " << firstLine);
+            // we must not return early with convert-to so that we clean up
+            // the session
+            if (!isConvertTo)
+            {
+                sendTextFrame("error: cmd=saveas kind=syntax");
+                return false;
+            }
+        }
+
+        std::string encodedWopiFilename;
+        if (!isConvertTo && !getTokenString(tokens[2], "filename", encodedWopiFilename))
+        {
+            LOG_ERR("Bad syntax for: " << firstLine);
+            sendTextFrame("error: cmd=saveas kind=syntax");
             return false;
         }
 
+        std::string url, wopiFilename;
+        Poco::URI::decode(encodedURL, url);
+        Poco::URI::decode(encodedWopiFilename, wopiFilename);
+
         // Save-as completed, inform the ClientSession.
-        const std::string filePrefix("file:///");
-        if (url.find(filePrefix) == 0)
+        Poco::URI resultURL(url);
+        if (resultURL.getScheme() == "file")
         {
+            std::string relative(resultURL.getPath());
+            if (relative.size() > 0 && relative[0] == '/')
+                relative = relative.substr(1);
+
             // Rewrite file:// URLs, as they are visible to the outside world.
-            const Path path(docBroker->getJailRoot(), url.substr(filePrefix.length()));
+            const Path path(docBroker->getJailRoot(), relative);
             if (Poco::File(path).exists())
             {
-                url = filePrefix + path.toString().substr(1);
+                resultURL.setPath(path.toString());
             }
             else
             {
                 // Blank for failure.
-                LOG_DBG("SaveAs produced no output, producing blank url.");
-                url.clear();
+                LOG_DBG("SaveAs produced no output in '" << path.toString() << "', producing blank url.");
+                resultURL.clear();
             }
         }
 
-        if (_saveAsSocket)
-        {
-            Poco::URI resultURL(url);
-            LOG_TRC("Save-as URL: " << resultURL.toString());
+        LOG_TRC("Save-as URL: " << resultURL.toString());
 
+        if (!isConvertTo)
+        {
+            // Normal SaveAs - save to Storage and log result.
+            if (resultURL.getScheme() == "file" && !resultURL.getPath().empty())
+            {
+                // this also sends the saveas: result
+                LOG_TRC("Save-as path: " << resultURL.getPath());
+                docBroker->saveAsToStorage(getId(), resultURL.getPath(), wopiFilename);
+            }
+            else
+                sendTextFrame("error: cmd=storage kind=savefailed");
+        }
+        else
+        {
+            // using the convert-to REST API
             // TODO: Send back error when there is no output.
             if (!resultURL.getPath().empty())
             {
@@ -613,7 +696,12 @@ bool ClientSession::handleKitToClientMessage(const char* buffer, const int lengt
                 std::string encodedFilePath;
                 Poco::URI::encode(resultURL.getPath(), "", encodedFilePath);
                 LOG_TRC("Sending file: " << encodedFilePath);
+
+                const std::string fileName = Poco::Path(resultURL.getPath()).getFileName();
                 Poco::Net::HTTPResponse response;
+                if (!fileName.empty())
+                    response.set("Content-Disposition", "attachment; filename=\"" + fileName + "\"");
+
                 HttpHelper::sendFile(_saveAsSocket, encodedFilePath, mimeType, response);
             }
 
@@ -739,17 +827,32 @@ bool ClientSession::forwardToClient(const std::shared_ptr<Message>& payload)
     return true;
 }
 
-std::string ClientSession::getAccessToken() const
+Authorization ClientSession::getAuthorization() const
 {
-    std::string accessToken;
     Poco::URI::QueryParameters queryParams = _uriPublic.getQueryParameters();
+
+    // prefer the access_token
     for (auto& param: queryParams)
     {
         if (param.first == "access_token")
-            return param.second;
+        {
+            std::string decodedToken;
+            Poco::URI::decode(param.second, decodedToken);
+            return Authorization(Authorization::Type::Token, decodedToken);
+        }
     }
 
-    return std::string();
+    for (auto& param: queryParams)
+    {
+        if (param.first == "access_header")
+        {
+            std::string decodedHeader;
+            Poco::URI::decode(param.second, decodedHeader);
+            return Authorization(Authorization::Type::Header, decodedHeader);
+        }
+    }
+
+    return Authorization();
 }
 
 void ClientSession::onDisconnect()
@@ -819,9 +922,19 @@ void ClientSession::dumpState(std::ostream& os)
     os << "\t\tisReadOnly: " << isReadOnly()
        << "\n\t\tisDocumentOwner: " << _isDocumentOwner
        << "\n\t\tisAttached: " << _isAttached
-       << "\n";
+       << "\n\t\tkeyEvents: " << _keyEvents;
+
+    std::shared_ptr<StreamSocket> socket = _socket.lock();
+    if (socket)
+    {
+        uint64_t sent, recv;
+        socket->getIOStats(sent, recv);
+        os << "\n\t\tsent/keystroke: " << (double)sent/_keyEvents << "bytes";
+    }
+
+    os << "\n";
     _senderQueue.dumpState(os);
+
 }
 
-
 /* vim:set shiftwidth=4 softtabstop=4 expandtab: */
